"""Contains all unit tests for the docs CLI."""

import os
from pathlib import Path
from unittest.mock import MagicMock, patch

from typer.testing import CliRunner

from voraus_pipeline_utils.cli.main import app


@patch.dict(
    os.environ,
    {
        "API_URL": "https://api.example.com",
        "PROJECT_NAME": "my-project",
        "PROJECT_VERSION": "1.0.0",
        "API_USER": "user",
        "API_TOKEN": "token",
    },
)
@patch("voraus_pipeline_utils.cli.docs.docs_upload_wrapper")
def test_docs_upload_success(docs_upload_wrapper_mock: MagicMock, cli_runner: CliRunner) -> None:
    result = cli_runner.invoke(app=app, args=["docs", "upload", "/path/to/docs"])
    assert result.exit_code == 0

    docs_upload_wrapper_mock.assert_called_once_with(
        build_dir=Path("/path/to/docs"),
        api_url="https://api.example.com",
        project_name="my-project",
        project_version="1.0.0",
        api_user="user",
        api_token="token",  # noqa: S106
    )



"""Contains all docker utils."""

import json
import logging
from pathlib import Path

from voraus_pipeline_utils.methods.build import sanitize_branch_name
from voraus_pipeline_utils.methods.environment import (
    get_branch_name,
    get_build_number,
    get_docker_image_name_from_env,
    get_docker_registry_from_env,
    get_docker_repositories_from_env,
    get_docker_tags_from_env,
    get_git_tag,
    is_ci,
)
from voraus_pipeline_utils.methods.shell import execute_command

_logger = logging.getLogger(__name__)


def get_tags_from_common_vars(
    tags: list[str] | None = None,
    registry: str | None = None,
    repositories: list[str] | None = None,
    image_name: str | None = None,
) -> list[str]:
    """Parses and returns the docker tags to use from the common environment variable.

    Args:
        tags: The docker tags provided by the CLI.
        registry: The docker registry to use.
        repositories: The docker repositories provided by the CLI.
        image_name: The docker image name provided by the CLI.

    Returns:
        The list of docker image tags to use.
    """
    if tags is not None:
        return tags
    image_name = image_name or get_docker_image_name_from_env()
    repositories = repositories or get_docker_repositories_from_env(raise_on_none=True)
    registry = registry or get_docker_registry_from_env()
    return get_docker_tags_from_env() or get_all_docker_tags(
        image_name=image_name, repositories=repositories, registry=registry
    )


def get_docker_image_tag(
    name: str,
    version_tag: str,
    registry: str,
    target_repo: str = "docker",
) -> str:
    """Constructs a valid docker tag from the given parameters.

    Args:
        name: The name of the image.
        version_tag: The version tag.
        registry: The docker registry to use.
        target_repo: The target repository. Defaults to "docker".

    Returns:
        The docker image tag.
    """
    return f"{registry}/{target_repo}/{name}:{version_tag}"


def get_all_docker_tags(  # noqa: PLR0913
    image_name: str | None,
    repositories: list[str],
    registry: str,
    branch_name: str | None = None,
    build_number: int | None = None,
    tag: str | None = None,
) -> list[str]:
    """Returns a list of all docker tags for the given values / environment.

    Args:
        image_name: The docker image name.
        repositories: A list of all repositories to publish to.
        registry: The docker registry to use.
        branch_name: The branch name. Defaults to None.
        build_number: The build number. Defaults to None.
        tag: The build tag. Defaults to None.

    Raises:
        ValueError: If an invalid combination of input parameters are provided.

    Returns:
        A list of docker images for the given values / environment.
    """
    if len(repositories) == 0:
        msg = "No repositories provided"
        raise ValueError(msg)
    branch_name = branch_name or get_branch_name()
    build_number = build_number or get_build_number()
    tag = tag or get_git_tag()
    version_tags: list[str] = []
    docker_tags: list[str] = []
    if image_name in ["", None]:
        msg = "Invalid image name"
        raise ValueError(msg)
    if branch_name and build_number is None:
        msg = "Build number must be provided when branch name is given"
        raise ValueError(msg)
    if tag:
        version_tags.append(tag)
        version_tags.append("latest")
    elif branch_name:
        sanitized_branch_name = sanitize_branch_name(branch_name=branch_name)
        version_tags.append(f"{sanitized_branch_name}-latest")
        version_tags.append(f"{sanitized_branch_name}-{build_number}")
    else:  # pragma: no cover
        msg = f"Invalid combination of branch_name={branch_name}, build_number={build_number} and tag={tag}"
        raise ValueError(msg)

    for version_tag in version_tags:
        for repository in repositories:
            docker_tags.append(  # noqa: PERF401
                get_docker_image_tag(
                    name=image_name,  # type:ignore [arg-type]  # The image_name can't be None here
                    version_tag=version_tag,
                    target_repo=repository,
                    registry=registry,
                )
            )

    return docker_tags


def docker_build_wrapper(  # noqa: PLR0913
    *,
    dockerfile: Path = Path("Dockerfile"),
    path: Path = Path(),
    tags: list[str] | None = None,
    pull: bool = True,
    use_jfrog: bool | None = None,
    build_args: dict[str, str] | None = None,
) -> None:
    """Wrapper for docker build. Optionally uses the JFrog CLI.

    Args:
        dockerfile: The path to the Dockerfile to use. Defaults to Path("Dockerfile").
        path: The build path. Defaults to Path(".").
        tags: The tags for the docker image. Defaults to None.
        pull: Whether to pull the base image. Defaults to True.
        use_jfrog: If True, the JFrog CLI will be used as additional wrapper.
            If executed in a CI environment, this defaults to True.
        build_args: Additional build arguments for docker.

    """
    build_args = build_args or {}
    tags = tags or []
    use_jfrog = use_jfrog if use_jfrog is not None else is_ci()
    execute_command(
        command=[
            *(["jf"] if use_jfrog else []),
            "docker",
            "build",
            *(["--pull"] if pull else []),
            *[f"-t {tag}" for tag in tags],
            *[f"--build-arg {key}={value}" for key, value in build_args.items()],
            "-f",
            str(dockerfile),
            str(path),
        ]
    )


def docker_push_wrapper(
    *,
    tags: list[str],
    use_jfrog: bool | None = None,
) -> None:
    """Wrapper for docker push. Optionally uses the JFrog CLI.

    Args:
        tags: The tags to push.
        use_jfrog: If True, the JFrog CLI will be used as additional wrapper.
            If executed in a CI environment, this defaults to True.

    """
    use_jfrog = use_jfrog if use_jfrog is not None else is_ci()
    for tag in tags:
        execute_command(command=[*(["jf"] if use_jfrog else []), "docker", "push", tag])


def docker_list_tags_wrapper(*, tags: list[str], file: Path | None = None) -> None:
    """Log and optionally write Docker tags to a file.

    Args:
        tags: List of Docker image tags to be logged and/or written to file.
        file: Optional path to write the tags as JSON.
    """
    _logger.info("The following tags have been generated from the environment variables and/or your input")
    for tag in tags:
        _logger.info("\t- %s", tag)
    if file is not None:
        file.parent.mkdir(parents=True, exist_ok=True)
        file.write_text(json.dumps(tags), encoding="utf-8")
        _logger.info("Tags have been written to file %s", file)


def docker_save_wrapper(*, tags: list[str], output: Path) -> None:
    """Saves the docker images to a file.

    Args:
        tags: List of Docker image tags to save to file.
        output: The output file.
    """
    _logger.info("Saving image(s) %s to file '%s'", tags, output)
    output.parent.mkdir(parents=True, exist_ok=True)
    execute_command(["docker", "save", "-o", str(output), *tags])



"""This module contains pytest fixtures."""

from pathlib import Path

import pytest
from typer.testing import CliRunner

from voraus_pipeline_utils.methods.shell import execute_command


@pytest.fixture(scope="session", name="resource_dir")
def resource_dir_fixture() -> Path:
    """Returns the path to the test resource directory.

    Returns:
        Path: The resource directory path.
    """
    return Path(__file__).parent.joinpath("resources")


@pytest.fixture(scope="session", name="cli_runner")
def cli_runner_fixture() -> CliRunner:
    """Returns a typer/click CliRunner with increased terminal width.

    Returns:
        The CliRunner object for testing the CLI.
    """
    return CliRunner(env={"COLUMNS": "160"})


@pytest.fixture(scope="session", name="dummy_images")
def dummy_images_fixture() -> list[str]:
    """Pulls some dummy docker images.

    Returns:
        The list of docker images that have been pulled.
    """
    dummy_images = ["alpine:latest", "alpine:edge"]
    for tag in dummy_images:
        execute_command(["docker", "pull", tag])
    return dummy_images



"""Contains all environment utils."""

import logging
import os
from collections.abc import Callable
from typing import Any, Literal, overload

from voraus_pipeline_utils.constants import (
    ENV_VAR_DOCKER_REGISTRY,
    ENV_VAR_DOCKER_TAGS,
    ENV_VAR_IMAGE_NAME,
    ENV_VAR_REPOSITORIES,
)

_logger = logging.getLogger(__name__)


def _get_env_or_none(env_var: str | tuple[str, ...], decode: Callable[[str], Any] | None = None) -> Any:  # noqa: ANN401
    """Get an environment variable value or None if not set.

    Supports checking multiple environment variable names (for fallback behavior) or a single variable.
    If multiple variables are provided and more than one is set, a warning is logged and the first
    non-null value is returned.

    Args:
        env_var: A single environment variable name or a tuple of variable names to check.
            If a tuple is provided, variables are checked in order and the first non-null value is returned.
        decode: An optional callable to transform the string value into another type.
            If None, the raw string value is returned.

    Returns:
        The environment variable value (optionally decoded), or None if the variable is not set or empty.
    """
    if isinstance(env_var, tuple):
        results = [_get_env_or_none(variable, decode=decode) for variable in env_var]

        non_null_results = [r for r in results if r is not None]
        result = None if len(non_null_results) == 0 else non_null_results[0]
        if len(non_null_results) > 1:
            _logger.warning(
                "Multiple environment variables from %s are set: %s. Using the first one.",
                env_var,
                non_null_results,
            )

    elif isinstance(env_var, str):
        if (value := os.environ.get(env_var)) not in ["", None]:
            if decode is not None:  # noqa: SIM108
                result = decode(value)  # type: ignore[arg-type]
            else:
                result = value
        else:
            result = None
    return result


def is_github() -> bool:
    """Tries to determine whether the current build is executed on GitHub actions or not.

    Returns:
        True if the current environment is a GitHub Action build, False otherwise.
    """
    github_actions = _get_env_or_none("GITHUB_ACTIONS")
    _logger.debug("GitHub Action build: %s", github_actions)
    return github_actions is not None


def is_jenkins() -> bool:
    """Tries to determine whether the current build is executed on a Jenkins or not.

    Returns:
        True if the current environment is on a Jenkins, False otherwise.
    """
    jenkins_url = _get_env_or_none("JENKINS_URL")
    _logger.debug("Jenkins URL is %s", jenkins_url)
    return jenkins_url is not None


def is_ci() -> bool:
    """Tries to determine whether the current build is executed in a CI pipeline.

    Returns:
        True if the current environment is a CI pipeline, False otherwise.
    """
    ci = _get_env_or_none("CI")
    _logger.debug("CI is %s", ci)
    return ci is not None


def get_git_tag() -> str | None:
    """Returns the tag name if building a tag, else None.

    Returns:
        The git tag from the environment, else None.
    """
    if is_github() and _get_env_or_none("GITHUB_REF_TYPE") == "tag":
        tag = _get_env_or_none("GITHUB_REF_NAME")
    else:
        tag = _get_env_or_none("TAG_NAME")

    _logger.debug("Tag is %s", tag)
    return tag


def get_build_number() -> int | None:
    """Returns the build number if available.

    Returns:
        The build number if set or None.
    """
    build_number = _get_env_or_none(("BUILD_NUMBER", "GITHUB_RUN_NUMBER"), decode=int)
    _logger.debug("Build number is %s", build_number)
    return build_number


def get_branch_name() -> str | None:
    """Returns the branch name.

    Returns:
        The branch name if set or None.
    """
    branch_name = _get_env_or_none(("BRANCH_NAME", "GITHUB_REF_NAME"))
    _logger.debug("Branch name is %s", branch_name)
    return branch_name


def get_change_target_branch_name() -> str | None:
    """Returns the branch name of the change target.

    Returns:
        The name of the branch the pull request points to if this is a pull request, None otherwise.
    """
    change_target = _get_env_or_none(("CHANGE_TARGET", "GITHUB_BASE_REF"))
    _logger.debug("Change target is %s", change_target)
    return change_target


def get_docker_tags_from_env() -> list[str] | None:
    """Returns the docker tag list from an optional environment variable.

    Returns:
        The docker tag list from the environment variable ``ENV_VAR_DOCKER_TAGS`` if existing, else None.
    """
    docker_tags = _get_env_or_none(ENV_VAR_DOCKER_TAGS, lambda tag_str: tag_str.split(","))
    _logger.debug("Docker tags are %s", docker_tags)
    return docker_tags


def get_docker_image_name_from_env() -> str | None:
    """Returns the docker image name from an optional environment variable.

    Returns:
        The docker image name from the environment variable ``ENV_VAR_IMAGE_NAME`` if existing, else None.
    """
    image_name = _get_env_or_none(ENV_VAR_IMAGE_NAME)
    _logger.debug("Docker image name is %s", image_name)
    return image_name


def get_docker_registry_from_env() -> str:
    """Returns the docker registry from an optional environment variable.

    Raises:
        ValueError: If the environment variable is not set.

    Returns:
        The docker registry from the environment variable ``ENV_VAR_DOCKER_REGISTRY``.
    """
    if (registry := _get_env_or_none(ENV_VAR_DOCKER_REGISTRY)) is None:
        error_msg = f"Unable to determine docker registry from environment variable {ENV_VAR_DOCKER_REGISTRY}"
        raise ValueError(error_msg)
    _logger.debug("Docker registry is %s", registry)
    return registry


@overload
def get_docker_repositories_from_env(*, raise_on_none: Literal[True]) -> list[str]: ...


@overload
def get_docker_repositories_from_env(*, raise_on_none: Literal[False]) -> list[str] | None: ...


def get_docker_repositories_from_env(*, raise_on_none: bool = False) -> list[str] | None:
    """Returns the docker repository list from an optional environment variable.

    Args:
        raise_on_none: If True, raises a ValueError if the environment variable is not set.

    Returns:
        The docker repository list from the environment variable ``ENV_VAR_REPOSITORIES`` if existing, else None.

    Raises:
        ValueError: If the environment variable is not set and ``raise_on_none`` is True.
    """
    repositories = _get_env_or_none(ENV_VAR_REPOSITORIES, lambda tag_str: tag_str.split(","))
    if not repositories and raise_on_none:
        error_msg = f"Unable to determine docker repositories from environment variable {ENV_VAR_REPOSITORIES}"
        raise ValueError(error_msg)
    _logger.debug("Docker repositories are %s", repositories)
    return repositories



"""Contains all tests for the docs_upload_wrapper function."""

import zipfile
from pathlib import Path
from unittest.mock import MagicMock, patch

from voraus_pipeline_utils.methods.docs import docs_upload_wrapper


@patch("voraus_pipeline_utils.methods.docs.TemporaryDirectory")
@patch("voraus_pipeline_utils.methods.docs.requests.post")
def test_docs_upload_wrapper(post_request_mock: MagicMock, temp_dir_class_mock: MagicMock, tmp_path: Path) -> None:
    build_dir = tmp_path / "build"
    build_dir.mkdir()
    (build_dir / "index.html").write_text("Hello")
    temp_dir_class_mock.return_value.__enter__.return_value = tmp_path
    docs_upload_wrapper(
        build_dir=build_dir,
        api_url="https://example.com/api/",
        project_name="my-project",
        project_version="1.0.0",
        api_user="user",
        api_token="token",  # noqa: S106
    )
    temp_dir_class_mock.assert_called_once()
    expected_docs_zip = tmp_path / "docs.zip"
    assert expected_docs_zip.is_file()
    with zipfile.ZipFile(expected_docs_zip, "r") as zip_handler:
        assert zip_handler.namelist() == ["index.html"]
        assert zip_handler.read("index.html").decode() == "Hello"
    post_request_mock.assert_called_once()



"""Contains all unit tests for the push CLI."""

from unittest.mock import MagicMock, patch

from typer.testing import CliRunner

from voraus_pipeline_utils.cli.main import app


@patch("voraus_pipeline_utils.cli.docker.get_tags_from_common_vars")
@patch("voraus_pipeline_utils.cli.docker.docker_push_wrapper")
def test_docker_push_success(
    docker_push_wrapper_mock: MagicMock, get_tags_from_common_vars_mock: MagicMock, cli_runner: CliRunner
) -> None:
    get_tags_from_common_vars_mock.return_value = ["A", "B"]
    result = cli_runner.invoke(app=app, args=["docker", "push"])
    assert result.exit_code == 0

    get_tags_from_common_vars_mock.assert_called_once_with(tags=None, repositories=None, image_name=None)
    docker_push_wrapper_mock.assert_called_once_with(tags=["A", "B"])


@patch("voraus_pipeline_utils.cli.docker.get_tags_from_common_vars")
@patch("voraus_pipeline_utils.cli.docker.docker_push_wrapper")
def test_docker_push_success_with_tags(
    docker_push_wrapper_mock: MagicMock, get_tags_from_common_vars_mock: MagicMock, cli_runner: CliRunner
) -> None:
    get_tags_from_common_vars_mock.return_value = ["A", "B"]
    result = cli_runner.invoke(app=app, args=["docker", "push", "--tag", "main-42", "--tag", "latest"])
    assert result.exit_code == 0

    get_tags_from_common_vars_mock.assert_called_once_with(
        tags=["main-42", "latest"], repositories=None, image_name=None
    )
    docker_push_wrapper_mock.assert_called_once_with(tags=["A", "B"])



"""Contains all tests for the docker_push_wrapper function."""

from unittest.mock import MagicMock, call, patch

import pytest

from voraus_pipeline_utils.methods.docker import docker_push_wrapper


@pytest.mark.parametrize("is_ci", [True, False])
@patch("voraus_pipeline_utils.methods.docker.is_ci")
@patch("voraus_pipeline_utils.methods.docker.execute_command")
def test_docker_push_wrapper(execute_command_mock: MagicMock, is_ci_mock: MagicMock, is_ci: bool) -> None:
    is_ci_mock.return_value = is_ci
    docker_push_wrapper(tags=["test:latest", "test:main-42"])
    execute_command_mock.assert_has_calls(
        [
            call(command=[*(["jf"] if is_ci else []), "docker", "push", "test:latest"]),
            call(command=[*(["jf"] if is_ci else []), "docker", "push", "test:main-42"]),
        ]
    )



"""Contains all constants."""

ENV_VAR_LOG_LEVEL = "VPU_LOG_LEVEL"
ENV_VAR_DOCKER_TAGS = "VPU_DOCKER_TAGS"
ENV_VAR_IMAGE_NAME = "VPU_DOCKER_IMAGE_NAME"
ENV_VAR_REPOSITORIES = "VPU_DOCKER_REPOSITORIES"
ENV_VAR_DOCKER_REGISTRY = "VPU_DOCKER_REGISTRY"



"""Contains all unit tests for build utils."""

import pytest

from voraus_pipeline_utils.methods.build import sanitize_branch_name


@pytest.mark.parametrize(
    ("branch_name", "expected_output"),
    [
        ("test%20branch", "test-branch"),
        ("special%25characters", "special-characters"),
        ("valid-branch_name.123", "valid-branch_name.123"),
        ("a%2Fb%5Cc", "a-b-c"),  # a/b\c becomes a-b-c
        ("%21%40%23%24%25%5E", "------"),  # !@#$%^ becomes ------
        ("branch%2Fname", "branch-name"),  # branch/name becomes branch-name
        ("branch%5Cname", "branch-name"),  # branch\name becomes branch-name
        (
            "multiple%2Fslashes%2Fand%5Cbackslashes",
            "multiple-slashes-and-backslashes",
        ),  # multiple/slashes/and\backslashes becomes multiple-slashes-and-backslashes
    ],
)
def test_sanitize_branch_name(branch_name: str, expected_output: str) -> None:
    assert sanitize_branch_name(branch_name) == expected_output



"""Contains all unit tests for the get_all_docker_tags method."""

import pytest

from voraus_pipeline_utils.methods.docker import get_all_docker_tags


@pytest.mark.parametrize("image_name", ["", None])
def test_get_all_docker_tags_invalid_image_name(image_name: str | None) -> None:
    with pytest.raises(ValueError, match="Invalid image name"):
        get_all_docker_tags(repositories=["docker"], image_name=image_name, registry="example.com")


def test_get_all_docker_tags_branch_name_missing_build_number() -> None:
    with pytest.raises(ValueError, match="Build number must be provided when branch name is given"):
        get_all_docker_tags(repositories=["docker"], image_name="test", branch_name="test", registry="example.com")


def test_get_all_docker_tags_branch_name_missing_repositories() -> None:
    with pytest.raises(ValueError, match="No repositories provided"):
        get_all_docker_tags(repositories=[], image_name="test", branch_name="test", registry="example.com")


def test_test_get_all_docker_tags_branch() -> None:
    assert get_all_docker_tags(
        repositories=["docker"],
        image_name="test",
        branch_name="test",
        build_number=1,
        registry="example.com",
    ) == [
        "example.com/docker/test:test-latest",
        "example.com/docker/test:test-1",
    ]


def test_test_get_all_docker_tags_tag() -> None:
    assert get_all_docker_tags(
        repositories=["docker"], image_name="test", tag="my-tag", registry="artifactory.vorausrobotik.com"
    ) == [
        "artifactory.vorausrobotik.com/docker/test:my-tag",
        "artifactory.vorausrobotik.com/docker/test:latest",
    ]


def test_test_get_all_docker_tags_branch_and_tag() -> None:
    assert get_all_docker_tags(
        repositories=["docker"],
        image_name="test",
        branch_name="whatever",
        build_number=1,
        tag="1.2.0",
        registry="artifactory.vorausrobotik.com",
    ) == [
        "artifactory.vorausrobotik.com/docker/test:1.2.0",
        "artifactory.vorausrobotik.com/docker/test:latest",
    ]



"""Contains all tests for the docker_build_wrapper function."""

from unittest.mock import MagicMock, patch

import pytest

from voraus_pipeline_utils.methods.docker import docker_build_wrapper


@pytest.mark.parametrize("is_ci", [True, False])
@patch("voraus_pipeline_utils.methods.docker.is_ci")
@patch("voraus_pipeline_utils.methods.docker.execute_command")
def test_docker_build_wrapper(execute_command_mock: MagicMock, is_ci_mock: MagicMock, is_ci: bool) -> None:
    is_ci_mock.return_value = is_ci
    docker_build_wrapper(tags=["test:latest", "test:main-42"])
    execute_command_mock.assert_called_once_with(
        command=[
            *(["jf"] if is_ci else []),
            "docker",
            "build",
            "--pull",
            "-t test:latest",
            "-t test:main-42",
            "-f",
            "Dockerfile",
            ".",
        ]
    )



"""Contains all unit tests for the protect_wheel_wrapper function."""

from pathlib import Path
from unittest.mock import MagicMock, patch

import pytest
import requests

from voraus_pipeline_utils.methods.protect import protect_wheel_wrapper


@patch("voraus_pipeline_utils.methods.protect.requests.post")
def test_protect_wheel_wrapper_success_replace(post_request_mock: MagicMock, tmp_path: Path) -> None:
    # Create a dummy wheel file
    wheel_file = tmp_path / "test_wheel.whl"
    wheel_file.write_bytes(b"dummy wheel content")

    # Mock the response
    mock_response = MagicMock()
    mock_response.content = b"encrypted wheel content"
    post_request_mock.return_value = mock_response

    # Call the function
    protect_wheel_wrapper(
        wheel_file=wheel_file,
        identifier="test-package-1",
        api_endpoint="https://protector.example.com",
        access_key_id="test_key",
        access_secret_key="test_secret",  # noqa: S106
        replace=True,
    )

    # Verify the request was made correctly
    post_request_mock.assert_called_once()
    call_args = post_request_mock.call_args
    assert call_args.args[0] == "https://protector.example.com/encrypt/python"
    assert call_args.kwargs["params"] == {"identifier": "test-package-1"}
    assert call_args.kwargs["headers"] == {"access_key_id": "test_key", "access_secret_key": "test_secret"}
    expected_timeout = 60
    assert call_args.kwargs["timeout"] == expected_timeout
    assert "files" in call_args.kwargs

    # Verify the file was replaced
    assert wheel_file.exists()
    assert wheel_file.read_bytes() == b"encrypted wheel content"


@patch("voraus_pipeline_utils.methods.protect.requests.post")
def test_protect_wheel_wrapper_success_no_replace(post_request_mock: MagicMock, tmp_path: Path) -> None:
    # Create a dummy wheel file
    wheel_file = tmp_path / "test_wheel.whl"
    original_content = b"dummy wheel content"
    wheel_file.write_bytes(original_content)

    # Mock the response
    mock_response = MagicMock()
    mock_response.content = b"encrypted wheel content"
    post_request_mock.return_value = mock_response

    # Call the function
    protect_wheel_wrapper(
        wheel_file=wheel_file,
        identifier="test-package-1",
        api_endpoint="https://protector.example.com",
        access_key_id="test_key",
        access_secret_key="test_secret",  # noqa: S106
        replace=False,
    )

    # Verify the request was made correctly
    post_request_mock.assert_called_once()

    # Verify the original file was NOT replaced
    assert wheel_file.exists()
    assert wheel_file.read_bytes() == original_content

    # Verify a new encrypted file was created
    encrypted_file = tmp_path / "test_wheel_encrypted.whl"
    assert encrypted_file.exists()
    assert encrypted_file.read_bytes() == b"encrypted wheel content"


@patch("voraus_pipeline_utils.methods.protect.requests.post")
def test_protect_wheel_wrapper_strips_trailing_slash(post_request_mock: MagicMock, tmp_path: Path) -> None:
    # Create a dummy wheel file
    wheel_file = tmp_path / "test_wheel.whl"
    wheel_file.write_bytes(b"dummy wheel content")

    # Mock the response
    mock_response = MagicMock()
    mock_response.content = b"encrypted wheel content"
    post_request_mock.return_value = mock_response

    # Call the function with trailing slash in endpoint
    protect_wheel_wrapper(
        wheel_file=wheel_file,
        identifier="test-package-1",
        api_endpoint="https://protector.example.com/",
        access_key_id="test_key",
        access_secret_key="test_secret",  # noqa: S106
        replace=True,
    )

    # Verify the URL was constructed correctly without double slashes
    call_args = post_request_mock.call_args
    assert call_args.args[0] == "https://protector.example.com/encrypt/python"


@patch("voraus_pipeline_utils.methods.protect.requests.post")
def test_protect_wheel_wrapper_request_failure(post_request_mock: MagicMock, tmp_path: Path) -> None:
    # Create a dummy wheel file
    wheel_file = tmp_path / "test_wheel.whl"
    wheel_file.write_bytes(b"dummy wheel content")

    # Mock the response to raise an HTTP error
    mock_response = MagicMock()
    mock_response.raise_for_status.side_effect = requests.HTTPError("API Error")
    post_request_mock.return_value = mock_response

    # Call the function and expect an exception
    with pytest.raises(requests.HTTPError, match="API Error"):
        protect_wheel_wrapper(
            wheel_file=wheel_file,
            identifier="test-package-1",
            api_endpoint="https://protector.example.com",
            access_key_id="test_key",
            access_secret_key="test_secret",  # noqa: S106
            replace=True,
        )


@patch("voraus_pipeline_utils.methods.protect.requests.post")
def test_protect_wheel_wrapper_timeout(post_request_mock: MagicMock, tmp_path: Path) -> None:
    # Create a dummy wheel file
    wheel_file = tmp_path / "test_wheel.whl"
    wheel_file.write_bytes(b"dummy wheel content")

    # Mock the response to timeout
    post_request_mock.side_effect = requests.Timeout("Request timed out")

    # Call the function and expect a timeout exception
    with pytest.raises(requests.Timeout, match="Request timed out"):
        protect_wheel_wrapper(
            wheel_file=wheel_file,
            identifier="test-package-1",
            api_endpoint="https://protector.example.com",
            access_key_id="test_key",
            access_secret_key="test_secret",  # noqa: S106
            replace=True,
        )


def test_protect_wheel_wrapper_file_not_found() -> None:
    # Test with a non-existent file
    wheel_file = Path("/nonexistent/path/test_wheel.whl")

    with pytest.raises(FileNotFoundError):
        protect_wheel_wrapper(
            wheel_file=wheel_file,
            identifier="test-package-1",
            api_endpoint="https://protector.example.com",
            access_key_id="test_key",
            access_secret_key="test_secret",  # noqa: S106
            replace=True,
        )



"""Contains all integration tests for the docker save CLI."""

import sys
from pathlib import Path
from unittest.mock import MagicMock, patch

import pytest
from typer.testing import CliRunner

from voraus_pipeline_utils.cli.main import app


@pytest.mark.skipif(sys.platform != "linux", reason="Only supported on linux because it requires docker")
@patch("voraus_pipeline_utils.cli.docker.get_tags_from_common_vars")
def test_docker_save_cli(
    get_tags_from_common_vars_mock: MagicMock, cli_runner: CliRunner, dummy_images: list[str], tmp_path: Path
) -> None:
    get_tags_from_common_vars_mock.return_value = dummy_images

    output_file = tmp_path / "archive.tar"

    result = cli_runner.invoke(app=app, args=["docker", "save", "--output", str(output_file)])
    assert result.exit_code == 0
    assert output_file.is_file()



"""Contains all unit tests for build utils."""

import os
from collections.abc import Callable
from typing import Any
from unittest.mock import patch

import pytest

from voraus_pipeline_utils.constants import (
    ENV_VAR_DOCKER_REGISTRY,
    ENV_VAR_DOCKER_TAGS,
    ENV_VAR_IMAGE_NAME,
    ENV_VAR_REPOSITORIES,
)
from voraus_pipeline_utils.methods.environment import (
    _get_env_or_none,
    get_branch_name,
    get_build_number,
    get_change_target_branch_name,
    get_docker_image_name_from_env,
    get_docker_registry_from_env,
    get_docker_repositories_from_env,
    get_docker_tags_from_env,
    get_git_tag,
    is_ci,
    is_github,
    is_jenkins,
)


@patch.dict(os.environ, clear=True)
def test___get_env_or_none_empty() -> None:
    assert _get_env_or_none(env_var="DOES_NOT_EXIST") is None


@patch.dict(os.environ, {"KEY": "VALUE"})
def test___get_env_or_none_without_decode() -> None:
    assert _get_env_or_none(env_var="KEY") == "VALUE"


@patch.dict(os.environ, {"KEY": "1"})
def test___get_env_or_none_with_decode() -> None:
    assert _get_env_or_none(env_var="KEY", decode=int) == 1


@patch.dict(os.environ, {"KEY": "VALUE"})
def test___get_env_or_none_with_decode_error() -> None:
    with pytest.raises(ValueError, match="invalid literal for int"):
        assert _get_env_or_none(env_var="KEY", decode=int)


@pytest.mark.parametrize(
    ("env", "expected_result", "expected_log"),
    [
        (
            {"KEY1": "VALUE1", "KEY2": "VALUE2"},
            "VALUE1",
            "Multiple environment variables from ('KEY1', 'KEY2') are set: ['VALUE1', 'VALUE2']. Using the first one.",
        ),
        ({"KEY1": "VALUE1", "KEY2": ""}, "VALUE1", None),
        ({"KEY1": "", "KEY2": "VALUE2"}, "VALUE2", None),
    ],
)
def test___get_env_or_none_multiple_keys(
    env: dict, expected_result: str | None, expected_log: str | None, caplog: pytest.LogCaptureFixture
) -> None:
    with patch.dict(os.environ, env), caplog.at_level("WARNING"):
        assert _get_env_or_none(env_var=("KEY1", "KEY2")) == expected_result

        if expected_log:
            assert expected_log in caplog.messages


@patch.dict(os.environ, {"JENKINS_URL": "http://jenkins.example.com"})
def test_is_jenkins() -> None:
    assert is_jenkins() is True


@patch.dict(os.environ, {"GITHUB_ACTIONS": "true"})
def test_is_github() -> None:
    assert is_github() is True


@patch.dict(os.environ, clear=True)
def test_is_not_jenkins() -> None:
    assert is_jenkins() is False


@patch.dict(os.environ, clear=True)
def test_is_not_github() -> None:
    assert is_github() is False


@patch.dict(os.environ, {"CI": "true"})
def test_is_ci() -> None:
    assert is_ci() is True


@patch.dict(os.environ, clear=True)
def test_is_not_ci() -> None:
    assert is_ci() is False


@pytest.mark.parametrize(
    ("function", "environment_dict", "expected_result"),
    [
        # Jenkins
        (get_git_tag, {"TAG_NAME": "TEST"}, "TEST"),
        (get_branch_name, {"BRANCH_NAME": "TEST"}, "TEST"),
        (get_build_number, {"BUILD_NUMBER": "42"}, 42),
        (get_change_target_branch_name, {"CHANGE_TARGET": "main"}, "main"),
        # GitHub Actions
        (
            get_git_tag,
            {
                "GITHUB_ACTIONS": "true",
                "GITHUB_REF_TYPE": "tag",
                "GITHUB_REF_NAME": "MockedTag",
            },
            "MockedTag",
        ),
        (get_branch_name, {"GITHUB_REF_NAME": "github-branch-name"}, "github-branch-name"),
        (get_build_number, {"GITHUB_RUN_NUMBER": "43"}, 43),
        (get_change_target_branch_name, {"GITHUB_BASE_REF": "main"}, "main"),
        # VPU variables
        (get_docker_image_name_from_env, {ENV_VAR_IMAGE_NAME: "example-image-name"}, "example-image-name"),
        (get_docker_repositories_from_env, {ENV_VAR_REPOSITORIES: "rep1"}, ["rep1"]),
        (get_docker_repositories_from_env, {ENV_VAR_REPOSITORIES: "rep1,rep2"}, ["rep1", "rep2"]),
        (get_docker_tags_from_env, {ENV_VAR_DOCKER_TAGS: "tag1"}, ["tag1"]),
        (get_docker_tags_from_env, {ENV_VAR_DOCKER_TAGS: "tag1,tag2"}, ["tag1", "tag2"]),
    ],
)
def test_get_environment_variables(function: Callable, environment_dict: dict, expected_result: Any) -> None:
    with patch.dict(os.environ, environment_dict):
        assert function() == expected_result

    # Test without value
    with patch.dict(os.environ, clear=True):
        assert function() is None


def test_get_docker_registry_from_env() -> None:
    with patch.dict(os.environ, {ENV_VAR_DOCKER_REGISTRY: "example.com"}):
        assert get_docker_registry_from_env() == "example.com"
    with (
        patch.dict(os.environ, clear=True),
        pytest.raises(
            ValueError, match="Unable to determine docker registry from environment variable VPU_DOCKER_REGISTRY"
        ),
    ):
        get_docker_registry_from_env()



"""This module defines the typer docker wrapper."""

from pathlib import Path
from typing import Annotated

import typer

from voraus_pipeline_utils.methods.docker import (
    docker_build_wrapper,
    docker_list_tags_wrapper,
    docker_push_wrapper,
    docker_save_wrapper,
    get_tags_from_common_vars,
)

app = typer.Typer(name="docker")


@app.command("build", help="Wrapper for docker build that optionally uses JFrog CLI and auto generates image tags")
def _cli_docker_build(
    image_name: Annotated[
        str | None,
        typer.Option(help="Docker image name. Ignored, if tag(s) are provided."),
    ] = None,
    tag: Annotated[
        list[str] | None,
        typer.Option(help="Docker image tag(s)"),
    ] = None,
    repository: Annotated[
        list[str] | None,
        typer.Option(help="Docker repository or repositories"),
    ] = None,
    registry: Annotated[
        str | None,
        typer.Option(help="Docker registry. Ignored, if tag(s) are provided."),
    ] = None,
    dockerfile: Annotated[
        Path,
        typer.Option("--file", "-f", help="Dockerfile to use for building the image"),
    ] = Path("Dockerfile"),
) -> None:
    tags = get_tags_from_common_vars(tags=tag, registry=registry, repositories=repository, image_name=image_name)
    docker_build_wrapper(tags=tags, dockerfile=dockerfile)


@app.command("push", help="Wrapper for docker push that optionally uses JFrog CLI and auto uses generated image tags")
def _cli_docker_push(
    image_name: Annotated[
        str | None,
        typer.Option(help="Docker image name. Ignored, if tag(s) are provided."),
    ] = None,
    tag: Annotated[
        list[str] | None,
        typer.Option(help="Docker image tag(s)"),
    ] = None,
    repository: Annotated[
        list[str] | None,
        typer.Option(help="Docker repository or repositories"),
    ] = None,
) -> None:
    tags = get_tags_from_common_vars(tags=tag, repositories=repository, image_name=image_name)
    docker_push_wrapper(tags=tags)


@app.command("list-tags", help="Lists the docker tags that are generated from the build environment")
def _cli_docker_tags(
    image_name: Annotated[
        str | None,
        typer.Option(help="Docker image name. Ignored, if tag(s) are provided."),
    ] = None,
    tag: Annotated[
        list[str] | None,
        typer.Option(help="Docker image tag(s)"),
    ] = None,
    repository: Annotated[
        list[str] | None,
        typer.Option(help="Docker repository or repositories"),
    ] = None,
    registry: Annotated[
        str | None,
        typer.Option(help="Docker registry. Ignored, if tag(s) are provided."),
    ] = None,
    file: Annotated[
        Path | None,
        typer.Option(help="Optional file to write the docker tags to (JSON format)."),
    ] = None,
) -> None:
    tags = get_tags_from_common_vars(tags=tag, registry=registry, repositories=repository, image_name=image_name)
    docker_list_tags_wrapper(tags=tags, file=file)


@app.command("save", help="Saves the docker images that are generated from the build environment to an archive")
def _cli_docker_save(
    output: Annotated[
        Path,
        typer.Option(help="The output file."),
    ],
    image_name: Annotated[
        str | None,
        typer.Option(help="Docker image name. Ignored, if tag(s) are provided."),
    ] = None,
    tag: Annotated[
        list[str] | None,
        typer.Option(help="Docker image tag(s)"),
    ] = None,
    repository: Annotated[
        list[str] | None,
        typer.Option(help="Docker repository or repositories"),
    ] = None,
    registry: Annotated[
        str | None,
        typer.Option(help="Docker registry. Ignored, if tag(s) are provided."),
    ] = None,
) -> None:
    tags = get_tags_from_common_vars(tags=tag, registry=registry, repositories=repository, image_name=image_name)
    docker_save_wrapper(tags=tags, output=output)



"""Contains all unit tests for the main CLI."""

import os
from unittest.mock import MagicMock, patch

from typer.testing import CliRunner

from voraus_pipeline_utils import get_app_name
from voraus_pipeline_utils.cli.main import app
from voraus_pipeline_utils.constants import ENV_VAR_LOG_LEVEL


@patch("voraus_pipeline_utils.cli.main.get_app_version")
def test_main_app(get_app_version_mock: MagicMock, cli_runner: CliRunner) -> None:
    get_app_version_mock.return_value = 42
    result = cli_runner.invoke(app, ["--version"])
    assert result.stdout == "42\n"


@patch("voraus_pipeline_utils.cli.docker.get_tags_from_common_vars")
@patch("voraus_pipeline_utils.cli.docker.docker_build_wrapper")
def test_log_level(_mock1: MagicMock, _mock2: MagicMock, cli_runner: CliRunner) -> None:  # noqa: PT019
    with patch.dict(os.environ, clear=True):
        result = cli_runner.invoke(app=app, args=["docker", "list-tags"])
        assert f"Using {get_app_name()}" in result.stdout
    with patch.dict(os.environ, {ENV_VAR_LOG_LEVEL: "WARNING"}):
        result = cli_runner.invoke(app=app, args=["docker", "list-tags"])
        assert f"Using {get_app_name()}" not in result.stdout



"""This module defines the typer docs wrapper."""

from pathlib import Path
from typing import Annotated

import typer

from voraus_pipeline_utils.methods.docs import docs_upload_wrapper

app = typer.Typer(name="docs")


@app.command("upload", help="Uploads HTML documentation to a vdoc instance")
def _cli_docs_upload(  # noqa: PLR0913
    *,
    build_dir: Annotated[
        Path,
        typer.Argument(
            help="Path to the directory containing the HTML documentation to upload.",
        ),
    ] = Path("docs/build/html"),
    project_name: Annotated[
        str,
        typer.Option(help="Name of the vdoc project to upload documentation to.", envvar="PROJECT_NAME"),
    ],
    project_version: Annotated[
        str,
        typer.Option(help="Version of the vdoc project to upload documentation to.", envvar="PROJECT_VERSION"),
    ],
    api_url: Annotated[
        str,
        typer.Option(help="API URL of the vdoc instance to upload documentation to.", envvar="API_URL"),
    ] = "https://docs.vorausrobotik.com/api",
    api_user: Annotated[
        str,
        typer.Option(help="API user for the vdoc instance.", envvar="API_USER"),
    ],
    api_token: Annotated[
        str,
        typer.Option(help="API token for the vdoc instance.", envvar="API_TOKEN"),
    ],
) -> None:  # noqa: disable=D103
    docs_upload_wrapper(
        build_dir=build_dir,
        api_url=api_url,
        api_user=api_user,
        api_token=api_token,
        project_name=project_name,
        project_version=project_version,
    )



"""Contains all tests for the get_docker_image_tag method."""

from voraus_pipeline_utils.methods.docker import get_docker_image_tag


def test_get_docker_image_tag() -> None:
    assert (
        get_docker_image_tag(name="my-name", version_tag="latest", target_repo="example", registry="example.com")
        == "example.com/example/my-name:latest"
    )



"""Contains all generic build utils."""

import re
import urllib.parse


def sanitize_branch_name(branch_name: str) -> str:
    """Sanitizes the given branch name by url-decoding it and replacing invalid characters with dashes.

    Args:
        branch_name: The input branch name.

    Returns:
        The sanitized branch name.
    """
    decoded_name = urllib.parse.unquote(branch_name, encoding="utf-8")
    return re.sub(r"[^a-zA-Z0-9-_\.]", "-", decoded_name)



"""Contains the main CLI entry point."""

import logging

import typer
from voraus_logging_lib.logging import LogLevel, configure_logger

from voraus_pipeline_utils import get_app_name, get_app_version
from voraus_pipeline_utils.cli.docker import app as docker_typer
from voraus_pipeline_utils.cli.docs import app as docs_typer
from voraus_pipeline_utils.cli.protect import app as protect_typer
from voraus_pipeline_utils.constants import ENV_VAR_LOG_LEVEL

_logger = logging.getLogger(__name__)


app = typer.Typer()

app.add_typer(docker_typer)
app.add_typer(docs_typer)
app.add_typer(protect_typer)


def print_version(do_print: bool) -> None:  # noqa: FBT001
    """Prints the version of the software.

    Args:
        do_print: If the version shall be printed.

    Raises:
        typer.Exit: After the version was printed.
    """
    if do_print:
        print(get_app_version())  # noqa: T201
        raise typer.Exit


@app.callback()
def _common(
    _: bool = typer.Option(  # noqa: FBT001
        False,  # noqa: FBT003
        "--version",
        callback=print_version,
        is_eager=True,
        help="Print the installed version of the software.",
    ),
    log_level: LogLevel = typer.Option(LogLevel.INFO, help="The log level", envvar=ENV_VAR_LOG_LEVEL),  # noqa: B008
) -> None:
    configure_logger(log_level=log_level.value)
    _logger.info("Using %s@v%s", get_app_name(), get_app_version())


typer_click_object = typer.main.get_command(app)

if __name__ == "__main__":  # pragma: no cover
    app()



"""Contains all unit tests for the build CLI."""

from pathlib import Path
from unittest.mock import MagicMock, patch

import pytest
from typer.testing import CliRunner

from voraus_pipeline_utils.cli.main import app


@pytest.mark.parametrize("export_file", [True, False])
@patch("voraus_pipeline_utils.cli.docker.get_tags_from_common_vars")
def test_list_docker_tags_success(
    get_tags_from_common_vars_mock: MagicMock, cli_runner: CliRunner, tmp_path: Path, export_file: bool
) -> None:
    get_tags_from_common_vars_mock.return_value = ["A", "B"]

    args = ["docker", "list-tags"]
    output_file = tmp_path / "tags.json"
    if export_file:
        args.extend(["--file", str(output_file)])

    result = cli_runner.invoke(app=app, args=args)
    assert result.exit_code == 0

    get_tags_from_common_vars_mock.assert_called_once_with(tags=None, registry=None, repositories=None, image_name=None)

    assert "The following tags have been generated" in result.stdout
    assert "- A" in result.stdout
    assert "- B" in result.stdout

    if export_file:
        assert output_file.exists()
        assert output_file.read_text() == '["A", "B"]'
        assert "Tags have been written to file" in result.stdout
        assert str(output_file) in result.stdout
    else:
        assert not output_file.exists()



"""Contains all shell helpers."""

import logging
import subprocess
from collections.abc import Callable
from pathlib import Path

_logger = logging.getLogger(__name__)


def execute_command(
    command: list[str],
    *,
    stdout_fn: Callable | None = print,
    no_raise: bool = False,
    cwd: Path | None = None,
) -> tuple[int, list[str]]:
    """Executes the given command and returns the result.

    Args:
        command: The command to execute.
        stdout_fn: The function to write the output to. Defaults to print.
        no_raise: If True, no exception is raised if the command Fails. Defaults to False.
        cwd: The working directory. Defaults to None.

    Raises:
        ValueError: If the stdout property cannot be assigned to the process.
        subprocess.CalledProcessError: If the command fails and _raise is True.

    Returns:
        A tuple containing the exit code and the stdout as list of the command.
    """
    command_str = " ".join(command)
    _logger.debug("Executing command '%s'", command_str)

    output: list[str] = []

    with subprocess.Popen(  # noqa: S602
        command_str,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,  # Redirect stderr to stdout
        text=True,  # To get strings instead of bytes
        bufsize=1,  # Line-buffered
        shell=True,
        cwd=cwd,
    ) as process:
        if process.stdout is None:
            msg = "Failed to set stdout for the subprocess."
            raise ValueError(msg)

        for output_line in process.stdout:
            _output_line = output_line.strip()
            output.append(_output_line)
            if stdout_fn is not None:
                stdout_fn(_output_line)

        # Wait for the process to complete
        process.wait()

        # Check the return code
        if process.returncode != 0 and not no_raise:
            raise subprocess.CalledProcessError(process.returncode, command)

    return process.returncode, output



"""This module defines the protect CLI commands for encrypting Python wheel files."""

from pathlib import Path
from typing import Annotated

import typer

from voraus_pipeline_utils.methods.protect import protect_wheel_wrapper

app = typer.Typer(name="protect")


@app.command("wheel")
def _cli_protect_wheel(  # noqa: PLR0913
    *,
    wheel_file: Annotated[
        Path,
        typer.Argument(
            help="Path to the wheel file to protect.",
        ),
    ],
    identifier: Annotated[
        str,
        typer.Option(help="The protection identifier"),
    ],
    replace: Annotated[
        bool,
        typer.Option(help="Whether to replace the original wheel or create a new file."),
    ] = True,
    api_endpoint: Annotated[
        str,
        typer.Option(help="The protector API endpoint.", envvar="PROTECTOR_API_ENDPOINT"),
    ] = "https://protector.vorausrobotik.com",
    access_key_id: Annotated[
        str,
        typer.Option(help="Access key ID for authentication.", envvar="PROTECTOR_ACCESS_KEY_ID", hide_input=True),
    ],
    access_secret_key: Annotated[
        str,
        typer.Option(
            help="Access secret key for authentication.", envvar="PROTECTOR_ACCESS_SECRET_KEY", hide_input=True
        ),
    ],
) -> None:
    r"""Encrypt Python wheel files using the voraus-software-protector service.

    This feature enables secure distribution of protected Python packages in CI/CD pipelines.
    The command encrypts wheel files via REST API. Credentials can be provided via environment
    variables or CLI options.

    Environment Variables:
    - PROTECTOR_ACCESS_KEY_ID: Access key ID for authentication
    - PROTECTOR_ACCESS_SECRET_KEY: Access secret key for authentication
    - PROTECTOR_API_ENDPOINT: API endpoint URL (optional, defaults to https://protector.vorausrobotik.com)

    Examples:
    Basic usage with environment variables::

        export PROTECTOR_ACCESS_KEY_ID="your-key-id"
        export PROTECTOR_ACCESS_SECRET_KEY="your-secret-key"
        vpu protect wheel dist/my_package-1.0.0-py3-none-any.whl --identifier=my-package-1

    Using explicit credentials::

        vpu protect wheel dist/my_package-1.0.0-py3-none-any.whl \\
            --identifier=my-package-1 \\
            --access-key-id=your-key-id \\
            --access-secret-key=your-secret-key

    Create encrypted copy without replacing original::

        vpu protect wheel dist/my_package-1.0.0-py3-none-any.whl \\
            --identifier=my-package-1 \\
            --no-replace

    Jenkins Pipeline Example::

        pipeline {
            agent any
            stages {
                stage('Protect Wheel') {
                    steps {
                        withCredentials([
                            usernamePassword(
                                credentialsId: 'protector-credentials',
                                usernameVariable: 'PROTECTOR_ACCESS_KEY_ID',
                                passwordVariable: 'PROTECTOR_ACCESS_SECRET_KEY'
                            )
                        ]) {
                            sh '''
                                vpu protect wheel dist/*.whl --identifier=my-package-1
                            '''
                        }
                    }
                }
            }
        }

    GitHub Actions Example::

        - name: Protect wheel
          env:
            PROTECTOR_ACCESS_KEY_ID: ${{ secrets.PROTECTOR_ACCESS_KEY_ID }}
            PROTECTOR_ACCESS_SECRET_KEY: ${{ secrets.PROTECTOR_ACCESS_SECRET_KEY }}
          run: |
            vpu protect wheel dist/*.whl --identifier=my-package-1

    """
    protect_wheel_wrapper(
        wheel_file=wheel_file,
        identifier=identifier,
        api_endpoint=api_endpoint,
        access_key_id=access_key_id,
        access_secret_key=access_secret_key,
        replace=replace,
    )



"""Contains all protection utils."""

import logging
from pathlib import Path

import requests

_logger = logging.getLogger(__name__)


def protect_wheel_wrapper(  # noqa: PLR0913
    *, wheel_file: Path, identifier: str, api_endpoint: str, access_key_id: str, access_secret_key: str, replace: bool
) -> None:
    """Protect a Python wheel file using the protector service.

    Args:
        wheel_file: Path to the wheel file to protect
        identifier: Protection identifier (e.g., 'dev-software-py-package-1')
        api_endpoint: The protector API endpoint
        access_key_id: Access key ID for authentication
        access_secret_key: Access secret key for authentication
        replace: If True, replace the original wheel file with the encrypted version
    """
    url = f"{api_endpoint.rstrip('/')}/encrypt/python"
    params = {"identifier": identifier}
    headers = {"access_key_id": access_key_id, "access_secret_key": access_secret_key}

    _logger.info("Sending wheel '%s' to '%s' for encryption", wheel_file, url)

    with wheel_file.open("rb") as file_handler:
        files = {"source": file_handler}
        response = requests.post(url, headers=headers, params=params, files=files, timeout=60)
        response.raise_for_status()

    output_file = wheel_file if replace else wheel_file.parent / f"{wheel_file.stem}_encrypted{wheel_file.suffix}"

    _logger.info("Writing protected wheel to file '%s'", output_file)

    with output_file.open("wb") as file_handler:
        file_handler.write(response.content)



"""Contains all integration tests for the protect wheel CLI."""

import os
from pathlib import Path
from unittest.mock import MagicMock, patch

from typer.testing import CliRunner

from voraus_pipeline_utils.cli.main import app


@patch.dict(
    os.environ,
    {
        "PROTECTOR_ACCESS_KEY_ID": "test_key_id",
        "PROTECTOR_ACCESS_SECRET_KEY": "test_secret_key",
    },
)
@patch("voraus_pipeline_utils.cli.protect.protect_wheel_wrapper")
def test_protect_wheel_success(protect_wheel_wrapper_mock: MagicMock, cli_runner: CliRunner) -> None:
    result = cli_runner.invoke(app=app, args=["protect", "wheel", "/path/to/wheel.whl", "--identifier=test-package-1"])
    assert result.exit_code == 0

    protect_wheel_wrapper_mock.assert_called_once_with(
        wheel_file=Path("/path/to/wheel.whl"),
        identifier="test-package-1",
        api_endpoint="https://protector.vorausrobotik.com",
        access_key_id="test_key_id",
        access_secret_key="test_secret_key",  # noqa: S106
        replace=True,
    )


@patch.dict(
    os.environ,
    {
        "PROTECTOR_ACCESS_KEY_ID": "test_key_id",
        "PROTECTOR_ACCESS_SECRET_KEY": "test_secret_key",
    },
)
@patch("voraus_pipeline_utils.cli.protect.protect_wheel_wrapper")
def test_protect_wheel_with_custom_endpoint(protect_wheel_wrapper_mock: MagicMock, cli_runner: CliRunner) -> None:
    result = cli_runner.invoke(
        app=app,
        args=[
            "protect",
            "wheel",
            "/path/to/wheel.whl",
            "--identifier=test-package-1",
            "--api-endpoint=https://custom.endpoint.com",
        ],
    )
    assert result.exit_code == 0

    protect_wheel_wrapper_mock.assert_called_once_with(
        wheel_file=Path("/path/to/wheel.whl"),
        identifier="test-package-1",
        api_endpoint="https://custom.endpoint.com",
        access_key_id="test_key_id",
        access_secret_key="test_secret_key",  # noqa: S106
        replace=True,
    )


@patch.dict(
    os.environ,
    {
        "PROTECTOR_ACCESS_KEY_ID": "test_key_id",
        "PROTECTOR_ACCESS_SECRET_KEY": "test_secret_key",
    },
)
@patch("voraus_pipeline_utils.cli.protect.protect_wheel_wrapper")
def test_protect_wheel_no_replace(protect_wheel_wrapper_mock: MagicMock, cli_runner: CliRunner) -> None:
    result = cli_runner.invoke(
        app=app,
        args=["protect", "wheel", "/path/to/wheel.whl", "--identifier=test-package-1", "--no-replace"],
    )
    assert result.exit_code == 0

    protect_wheel_wrapper_mock.assert_called_once_with(
        wheel_file=Path("/path/to/wheel.whl"),
        identifier="test-package-1",
        api_endpoint="https://protector.vorausrobotik.com",
        access_key_id="test_key_id",
        access_secret_key="test_secret_key",  # noqa: S106
        replace=False,
    )


@patch("voraus_pipeline_utils.cli.protect.protect_wheel_wrapper")
def test_protect_wheel_with_explicit_credentials(protect_wheel_wrapper_mock: MagicMock, cli_runner: CliRunner) -> None:
    result = cli_runner.invoke(
        app=app,
        args=[
            "protect",
            "wheel",
            "/path/to/wheel.whl",
            "--identifier=test-package-1",
            "--access-key-id=explicit_key",
            "--access-secret-key=explicit_secret",
        ],
    )
    assert result.exit_code == 0

    protect_wheel_wrapper_mock.assert_called_once_with(
        wheel_file=Path("/path/to/wheel.whl"),
        identifier="test-package-1",
        api_endpoint="https://protector.vorausrobotik.com",
        access_key_id="explicit_key",
        access_secret_key="explicit_secret",  # noqa: S106
        replace=True,
    )


@patch.dict(
    os.environ,
    {
        "PROTECTOR_API_ENDPOINT": "https://env-endpoint.example.com",
        "PROTECTOR_ACCESS_KEY_ID": "test_key_id",
        "PROTECTOR_ACCESS_SECRET_KEY": "test_secret_key",
    },
)
@patch("voraus_pipeline_utils.cli.protect.protect_wheel_wrapper")
def test_protect_wheel_with_endpoint_from_env(protect_wheel_wrapper_mock: MagicMock, cli_runner: CliRunner) -> None:
    result = cli_runner.invoke(app=app, args=["protect", "wheel", "/path/to/wheel.whl", "--identifier=test-package-1"])
    assert result.exit_code == 0

    protect_wheel_wrapper_mock.assert_called_once_with(
        wheel_file=Path("/path/to/wheel.whl"),
        identifier="test-package-1",
        api_endpoint="https://env-endpoint.example.com",
        access_key_id="test_key_id",
        access_secret_key="test_secret_key",  # noqa: S106
        replace=True,
    )



"""Contains all docs utils."""

import zipfile
from pathlib import Path
from tempfile import TemporaryDirectory

import requests


def docs_upload_wrapper(  # noqa: PLR0913
    *,
    build_dir: Path,
    api_url: str,
    project_name: str,
    project_version: str,
    api_user: str,
    api_token: str,
) -> None:
    """Uploads documentation to a docs (vdoc) instance.

    Args:
        build_dir: The directory containing the HTML documentation to upload.
        project_name: The name of the project to upload documentation to.
        project_version: The version of the project to upload documentation to.
        api_user: The API user for the docs instance.
        api_url: The URL of the docs instance to upload documentation to.
        api_token: The API token for the docs instance.

    Raises:
        FileNotFoundError: If the index.html file is not found in the build directory.

    """
    if not (build_dir / "index.html").is_file():
        msg = f"Index file not found in {build_dir}"
        raise FileNotFoundError(msg)
    with TemporaryDirectory() as tmp_dir:
        zip_file_path = Path(tmp_dir) / "docs.zip"
        with zipfile.ZipFile(zip_file_path, "w", zipfile.ZIP_DEFLATED) as zip_handler:
            for file in build_dir.rglob("*"):
                zip_handler.write(file, arcname=file.relative_to(build_dir))

        response = requests.post(
            f"{api_url.rstrip('/')}/projects/{project_name}/versions/{project_version}",
            auth=(api_user, api_token),
            files={"file": (zip_file_path.name, zip_file_path.read_bytes(), "application/zip")},
            timeout=60,
        )
        response.raise_for_status()



"""Contains all tests for the get_tags_from_common_vars function."""

import os
from unittest.mock import MagicMock, patch

from voraus_pipeline_utils.constants import ENV_VAR_DOCKER_REGISTRY
from voraus_pipeline_utils.methods.docker import get_tags_from_common_vars


@patch("voraus_pipeline_utils.methods.docker.get_all_docker_tags")
@patch("voraus_pipeline_utils.methods.docker.get_docker_tags_from_env")
@patch("voraus_pipeline_utils.methods.docker.get_docker_repositories_from_env")
@patch("voraus_pipeline_utils.methods.docker.get_docker_image_name_from_env")
def test_tags_provided_directly(
    mock_get_docker_image_name_from_env: MagicMock,
    mock_get_docker_repositories_from_env: MagicMock,
    mock_get_docker_tags_from_env: MagicMock,
    mock_get_all_docker_tags: MagicMock,
) -> None:
    # Scenario: Tags are provided directly, so they should be returned as-is.
    tags = ["latest", "1.0"]
    result = get_tags_from_common_vars(tags=tags, repositories=None, image_name=None, registry="example.com")
    assert result == tags
    mock_get_all_docker_tags.assert_not_called()
    mock_get_docker_tags_from_env.assert_not_called()
    mock_get_docker_repositories_from_env.assert_not_called()
    mock_get_docker_image_name_from_env.assert_not_called()


@patch("voraus_pipeline_utils.methods.docker.get_all_docker_tags")
@patch("voraus_pipeline_utils.methods.docker.get_docker_tags_from_env")
@patch("voraus_pipeline_utils.methods.docker.get_docker_repositories_from_env")
@patch("voraus_pipeline_utils.methods.docker.get_docker_image_name_from_env")
def test_tags_not_provided_image_name_provided(
    mock_get_docker_image_name_from_env: MagicMock,
    mock_get_docker_repositories_from_env: MagicMock,
    mock_get_docker_tags_from_env: MagicMock,
    mock_get_all_docker_tags: MagicMock,
) -> None:
    # Scenario: Tags are not provided, but image name is provided. Environment variables are not set.
    image_name = "test-image"
    repositories = ["repo1", "repo2"]
    generated_tags = ["generated-tag1", "generated-tag2"]

    mock_get_all_docker_tags.return_value = generated_tags
    mock_get_docker_tags_from_env.return_value = None

    result = get_tags_from_common_vars(
        tags=None, repositories=repositories, image_name=image_name, registry="example.com"
    )
    assert result == generated_tags
    mock_get_all_docker_tags.assert_called_once_with(
        image_name=image_name, repositories=repositories, registry="example.com"
    )
    mock_get_docker_tags_from_env.assert_called_once()

    mock_get_docker_image_name_from_env.assert_not_called()
    mock_get_docker_repositories_from_env.assert_not_called()


@patch("voraus_pipeline_utils.methods.docker.get_all_docker_tags")
@patch("voraus_pipeline_utils.methods.docker.get_docker_tags_from_env")
@patch("voraus_pipeline_utils.methods.docker.get_docker_repositories_from_env")
@patch("voraus_pipeline_utils.methods.docker.get_docker_image_name_from_env")
def test_tags_fetched_from_environment(
    mock_get_docker_image_name_from_env: MagicMock,
    mock_get_docker_repositories_from_env: MagicMock,
    mock_get_docker_tags_from_env: MagicMock,
    mock_get_all_docker_tags: MagicMock,
) -> None:
    # Scenario: Tags and image name are not provided, should fetch from environment
    image_name = "env-image"
    repositories = ["env-repo1", "env-repo2"]
    env_tags = ["env-tag1", "env-tag2"]

    mock_get_docker_image_name_from_env.return_value = image_name
    mock_get_docker_repositories_from_env.return_value = repositories
    mock_get_docker_tags_from_env.return_value = env_tags

    result = get_tags_from_common_vars(tags=None, repositories=None, image_name=None, registry="example.com")
    assert result == env_tags
    mock_get_docker_image_name_from_env.assert_called_once()
    mock_get_docker_repositories_from_env.assert_called_once()
    mock_get_docker_tags_from_env.assert_called_once()
    mock_get_all_docker_tags.assert_not_called()


@patch.dict(os.environ, {ENV_VAR_DOCKER_REGISTRY: "example.com"})
@patch("voraus_pipeline_utils.methods.docker.get_all_docker_tags")
@patch("voraus_pipeline_utils.methods.docker.get_docker_tags_from_env")
def test_tags_fetched_from_get_all_docker_tags(
    mock_get_docker_tags_from_env: MagicMock,
    mock_get_all_docker_tags: MagicMock,
) -> None:
    # Scenario: Tags are not provided, and get_all_docker_tags is used to generate tags
    image_name = "test-image"
    repositories = ["repo1", "repo2"]
    generated_tags = ["generated-tag1", "generated-tag2"]

    mock_get_docker_tags_from_env.return_value = None
    mock_get_all_docker_tags.return_value = generated_tags

    result = get_tags_from_common_vars(tags=None, repositories=repositories, image_name=image_name)
    assert result == generated_tags
    mock_get_all_docker_tags.assert_called_once_with(
        image_name=image_name, repositories=repositories, registry="example.com"
    )
    mock_get_docker_tags_from_env.assert_called_once()



"""Contains all application utility functions."""

import pytest

from voraus_pipeline_utils import get_app_name, get_app_version


def test_get_app_name() -> None:
    assert get_app_name() == "voraus-pipeline-utils"


def test_get_app_version(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr("voraus_pipeline_utils.__version__", "42.0.0")
    assert get_app_version() == "42.0.0"



"""Uses pip-licenses to determine the licenses of all installed packages and generates a report for the docs."""

import json
import os
import subprocess
from pathlib import Path
from tempfile import TemporaryDirectory

import jinja2

if __name__ == "__main__":
    DOCS_PATH = Path("docs")
    JINJA_ENVIRONMENT = jinja2.Environment(
        loader=jinja2.FileSystemLoader(DOCS_PATH / "_templates")
    )

    with TemporaryDirectory() as tmp_dir:
        tmp_path = Path(tmp_dir)
        licenses_file = tmp_path / "ossLicenses.json"

        subprocess.check_call(
            [
                "pip-licenses",
                "--with-authors",
                "--with-system",
                "--with-urls",
                "--with-description",
                "--with-license-file",
                "--with-notice-file",
                "--format",
                "json",
                "--ignore-packages",
                "voraus_pipeline_utils",
                "--output-file",
                licenses_file.as_posix(),
            ]
        )

        licenses: list[dict[str, str | None]] = json.loads(
            licenses_file.read_text(encoding="utf-8")
        )

        for license_ in licenses:
            for key in ["LicenseFile", "NoticeFile"]:
                file = license_[key]

                if file is not None and file != "UNKNOWN":
                    license_[key] = os.path.relpath(file, DOCS_PATH.as_posix())
                else:
                    license_[key] = None

        (DOCS_PATH / "license_compliance.rst").write_text(
            JINJA_ENVIRONMENT.get_template("license_compliance.rst.j2").render(
                licenses=licenses
            )
        )



"""A library containing pipeline utils for both, local usage and usage in Jenkins."""

from importlib_metadata import PackageNotFoundError as _PackageNotFoundError
from importlib_metadata import version as _version

__module_name__ = "voraus_pipeline_utils"

try:  # pragma: no cover
    __version__ = _version(__module_name__)
except _PackageNotFoundError as error:  # pragma: no cover
    msg = (
        f"Unable to determine version of package '{__module_name__}'. "
        "If you are on a local development system, use 'pip install -e .[dev]' in order to install the package. "
        "If you are on a productive system, this shouldn't happen. Please report a bug."
    )
    raise ModuleNotFoundError(msg) from error


def get_app_name() -> str:
    """Returns the human-readable and prettified name of the application.

    Returns:
        The name of the application.
    """
    return __module_name__.replace("_", "-")


def get_app_version() -> str:
    """Returns the version of the application.

    Returns:
        The version of the application.
    """
    return __version__



"""Contains all unit tests for the execute command helper."""

import subprocess
from io import StringIO
from pathlib import Path
from unittest.mock import MagicMock, patch

import pytest

from voraus_pipeline_utils.methods.shell import execute_command


def test_execute_command_success() -> None:
    exit_code, output = execute_command(["echo", "'Hello pytest!'"])
    assert exit_code == 0
    assert output == ["Hello pytest!"]


def test_execute_command_failure() -> None:
    with pytest.raises(subprocess.CalledProcessError):
        execute_command(["false"])


def test_execute_command_failure_no_raise() -> None:
    exit_code, output = execute_command(["false"], no_raise=True)
    assert exit_code == 1
    assert not output


def test_execute_command_stdout_fn_default(capsys: pytest.CaptureFixture) -> None:
    execute_command(["echo", "'Hello pytest!'"])
    assert capsys.readouterr().out == "Hello pytest!\n"


def test_execute_command_stdout_fn_default_multiline() -> None:
    _, output = execute_command(["echo 'Hello\nHello\nHello\nHello\nHello'"])
    assert output == ["Hello"] * 5


def test_execute_command_stdout_fn_overwrite(capsys: pytest.CaptureFixture) -> None:
    out = StringIO()
    execute_command(["echo", "'Hello pytest!'"], stdout_fn=out.write)
    assert capsys.readouterr().out == ""
    out.seek(0)
    assert out.read() == "Hello pytest!"


def test_execute_command_cwd(tmp_path: Path) -> None:
    (tmp_path / "test.txt").touch()
    _, output = execute_command(["ls"], cwd=tmp_path)
    assert output == ["test.txt"]


def test_execute_command_stdout_fn_none(capsys: pytest.CaptureFixture) -> None:
    _, output = execute_command(["echo", "'Hello pytest!'"], stdout_fn=None)
    assert capsys.readouterr().out == ""
    assert output == ["Hello pytest!"]


@patch("voraus_pipeline_utils.methods.shell.subprocess.Popen")
def test_execute_command_process_stdout_is_none(popen_mock: MagicMock) -> None:
    process_mock = MagicMock()
    process_mock.__enter__.return_value.stdout = None
    popen_mock.return_value = process_mock
    with pytest.raises(ValueError, match="Failed to set stdout for the subprocess"):
        execute_command(["echo", "'Hello pytest!'"])
